@@ -16,7 +16,9 @@ from io import BytesIO
 from ssl import CERT_NONE, CERT_OPTIONAL, CERT_REQUIRED
 
 import grpc
+from grpc._channel import _InactiveRpcError
 
+from e6data_python_connector.cluster_manager import ClusterManager
 from e6data_python_connector.common import DBAPITypeObject, ParamEscaper, DBAPICursor
 from e6data_python_connector.constants import *
 from e6data_python_connector.datainputstream import get_query_columns_info, read_rows_from_chunk
@@ -51,14 +53,35 @@ def _parse_timestamp(value):
             value = datetime.datetime.strptime(value, format)
         else:
             raise Exception(
-                'Cannot convert "{}" into a datetime'.format(value))
+                'Cannot convert "{}" into a datetime'.format(value)
+            )
     else:
         value = None
     return value
 
 
-TYPES_CONVERTER = {"DECIMAL_TYPE": Decimal,
-                   "TIMESTAMP_TYPE": _parse_timestamp}
+TYPES_CONVERTER = {
+    "DECIMAL_TYPE": Decimal,
+    "TIMESTAMP_TYPE": _parse_timestamp
+}
+
+
+def re_auth(func):
+    def wrapper(self, *args, **kwargs):
+        try:
+            return func(self, *args, **kwargs)
+        except _InactiveRpcError as e:
+            print(f'RE_AUTH: Function Name: {func}')
+            print(f'RE_AUTH: Error Found {e}')
+            if e.code() == grpc.StatusCode.INTERNAL and 'Access denied' in e.details():
+                print('RE_AUTH: Initialising re-authentication.')
+                self.connection.get_re_authenticate_session_id()
+                print(f'RE_AUTH: Re-auth successful.')
+                return func(self, *args, **kwargs)
+            else:
+                raise e
+
+    return wrapper
 
 
 class HiveParamEscaper(ParamEscaper):
@@ -114,6 +137,7 @@ class Connection(object):
             database: str = None,
             cluster_uuid: str = None,
             secure: bool = False,
+            auto_resume: bool = False,
             grpc_options: dict = None,
     ):
         """
@@ -137,6 +161,8 @@ class Connection(object):
                 Cluster's uuid
             secure: bool, Optional
                 Flag to use a secure channel for data transfer
+            auto_resume: bool, Optional
+                Flag to enable auto resume of the cluster before the query execution
             grpc_options: dict, Optional
                 Specify gRPC configuration
                 - keepalive_timeout_ms: This parameter defines the time, in milliseconds, that a gRPC connection should remain idle before sending a keepalive ping to check if the connection is still alive.
@@ -160,12 +186,14 @@ class Connection(object):
 
         self.catalog_name = catalog
 
+        self._auto_resume = auto_resume
+
         self._keepalive_timeout_ms = 900000
         self._max_receive_message_length = -1
         self._max_send_message_length = 300 * 1024 * 1024  # mb
         self.grpc_prepare_timeout = 10 * 60  # 10 minutes
 
-        if type(grpc_options) == dict:
+        if isinstance(grpc_options, dict):
             self._keepalive_timeout_ms = grpc_options.get('keepalive_timeout_ms') or self._keepalive_timeout_ms
             self._max_receive_message_length = grpc_options.get(
                 'max_receive_message_length') or self._max_receive_message_length
@@ -195,6 +223,10 @@ class Connection(object):
             )
         self._client = e6x_engine_pb2_grpc.QueryEngineServiceStub(self._channel)
 
+    def get_re_authenticate_session_id(self):
+        self._session_id = None
+        return self.get_session_id
+
     @property
     def get_session_id(self):
         """
@@ -213,6 +245,33 @@ class Connection(object):
                 self._session_id = authenticate_response.sessionId
                 if not self._session_id:
                     raise ValueError("Invalid credentials.")
+            except _InactiveRpcError as e:
+                if self._auto_resume:
+                    if e.code() == grpc.StatusCode.UNAVAILABLE and 'status: 503' in e.details():
+                        status = ClusterManager(
+                            host=self._host,
+                            port=self._port,
+                            user=self.__username,
+                            password=self.__password,
+                            secure_channel=self._secure_channel,
+                            cluster_uuid=self.cluster_uuid
+                        ).resume()
+                        if status:
+                            authenticate_request = e6x_engine_pb2.AuthenticateRequest(
+                                user=self.__username,
+                                password=self.__password
+                            )
+                            authenticate_response = self._client.authenticate(
+                                authenticate_request,
+                                metadata=_get_grpc_header(cluster=self.cluster_uuid)
+                            )
+                            self._session_id = authenticate_response.sessionId
+                        else:
+                            raise e
+                    else:
+                        raise e
+                else:
+                    raise e
             except Exception as e:
                 self._channel.close()
                 raise e
@@ -463,6 +522,7 @@ class Cursor(DBAPICursor):
         )
         return self.connection.client.status(status_request, metadata=self.metadata)
 
+    @re_auth
     def execute(self, operation, parameters=None, **kwargs):
         """Prepare and execute a database operation (query or command).
         Return values are not defined.
@@ -518,6 +578,7 @@ class Cursor(DBAPICursor):
 
             self._query_id = prepare_statement_response.queryId
             self._engine_ip = prepare_statement_response.engineIP
+
             execute_statement_request = e6x_engine_pb2.ExecuteStatementV2Request(
                 engineIP=self._engine_ip,
                 sessionId=self.connection.get_session_id,
@@ -602,7 +663,6 @@ class Cursor(DBAPICursor):
         return self._fetch_all()
 
     def fetchmany(self, size: int = None):
-        # _logger.info("fetching all from overriden method")
         if size is None:
             size = self.arraysize
         if self._data is None:
